VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "cPalette"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Attribute VB_Description = "Octree logic"
' ========================================================================
' cPalette.cls
' Adapted from code by Steve McMahon at:
' http://vbaccelerator.com
'
' The Octree Colour Quantisation Code (CreateOptimal) was written by
' Brian Schimpf Copyright  1999 Brian Schimpf
'
' Modified by: Stephen Lebans
' Stephen@lebans.com
' The above statements of copyright and credit are in no way
' to be construed that either party has given permission for
' the use of their code in this project.
'
' ========================================================================

Option Explicit

Private Declare Sub CopyMemory Lib "kernel32" Alias "RtlMoveMemory" ( _
    ByVal lpvDest As Long, lpvSource As Any, ByVal cbCopy As Long)

Private Declare Sub apiCopyMemory Lib "kernel32" Alias "RtlMoveMemory" ( _
   lpvDest As Any, lpvSource As Any, ByVal cbCopy As Long)


Private Declare Function VarPtrArray Lib "msvbvm50.dll" Alias _
   "VarPtr" (Ptr() As Any) As Long

Private Type SAFEARRAYBOUND
    cElements As Long
    lLbound As Long
End Type

Private Type SAFEARRAY2D
    cDims As Integer
    fFeatures As Integer
    cbElements As Long
    cLocks As Long
    pvData As Long
    Bounds(0 To 1) As SAFEARRAYBOUND
End Type


Private Type tNode                              'Octree node struct
    bIsLeaf As Boolean                          'Leaf flag
    bAddedReduce As Boolean                     'Linked list flag
    vRed As Long                                'Red Value
    vGreen As Long                              'Green Value
    vBlue As Long                               'Blue Value
    cClrs As Long                               'Pixel count
    iChildren(0 To 1, 0 To 1, 0 To 1) As Long   'Child pointers
    iNext As Long                               'Next reducable node
End Type

Private Type RGBQUAD
   Red As Byte
   Green As Byte
   Blue As Byte
   Reserved As Byte
End Type

' Pixel values are stored in memory as BGR
Private Type BGRQUAD
   Blue As Byte
   Green As Byte
   Red As Byte
   Reserved As Byte
End Type

' Module level constants
' The human eye is sensitive to certain areas
' of the color spectrum than others.
Private Const EyeSenBlue = 11
Private Const EyeSenGreen = 59
Private Const EyeSenRed = 30

' Module level vars
Private aNodes() As tNode, cNodes As Long
Private nDepth As Byte, TopGarbage As Long
Private cClr As Long, aReduce(1 To 8) As Long
Private cOps As Long

Private m_tPal() As RGBQUAD
' Number of entries in our m_tPal() RGBQUAD array
Private m_iPalette As Long

' Need to use internal prop - get rid of this
Private lngBytesPerScanLine As Long


Public Sub Add(ByVal Red As Byte, ByVal Green As Byte, ByVal Blue As Byte)
    m_iPalette = m_iPalette + 1
    ReDim Preserve m_tPal(1 To m_iPalette) As RGBQUAD
    m_tPal(m_iPalette).Red = Red
    m_tPal(m_iPalette).Green = Green
    m_tPal(m_iPalette).Blue = Blue
End Sub

Public Property Get ClosestIndex( _
      ByVal Red As Long, _
      ByVal Green As Long, _
      ByVal Blue As Long _
   ) As Long


' * NOTE SL
' This is an algorithm I converted from a
' C++ example I found on the Web. No author was mentioned.
Dim i As Long, d As Long, p As Long
Dim Min As Long, r As Long, g As Long, b As Long

' Init our vars
i = 0
p = 0
Min = &H7FFFFFFF
        
For i = 1 To m_iPalette
    With m_tPal(i)
        b = (Blue - .Blue) '* EyeSenBlue
        b = b * b
        g = (Green - .Green) ' * EyeSenGreen
        g = g * g
        r = (Red - .Red) '* EyeSenRed
        r = r * r
        d = r + g + b
        If (d < Min) Then
            Min = d
            p = i
        End If
    End With
Next
ClosestIndex = p
End Property

Public Property Get Count() As Long
' Number of colours in this palette:
   Count = m_iPalette
End Property
Public Property Get Red(ByVal iIndex) As Byte
' Red value for palette index
   Red = m_tPal(iIndex).Red
End Property
Public Property Get Green(ByVal iIndex) As Byte
' Green value for palette index
   Green = m_tPal(iIndex).Green
End Property
Public Property Get Blue(ByVal iIndex) As Byte
' Blue value for palette index
   Blue = m_tPal(iIndex).Blue
End Property

'***************************************************************
'CreateOptimalPalette--Creates an optimal palette with the
'  specified number of colors using octree quantisation
'
'Inputs:
'  cDib
'  DIBSection for which palette is made
'
'  nLevels
'  Number of levels of color to create
'
'  PalSize
'  Size of palette to make
'
'Date Created:7/12/1999          Brian Schimpf
'
'Modifications:                               Date:
'***************************************************************
Public Sub CreateOptimal( _
      ByRef cDIB As cDIBSection, ByRef cDIBSave As cDIBSectionSave, _
      Optional ByVal PalSize As Long = 256, _
      Optional ByVal nLevels As Long = 8 _
   )
Dim tSA As SAFEARRAY2D, aDib() As Byte
Dim x As Long, y As Long
Dim xMax As Long, yMax As Long
Dim sTime As Long, pPal As Double
Dim Inc As Double

Dim lngBytesPerScanLine As Long
Dim lpbitsptr As Long

' Temp RGB values
Dim bgr As BGRQUAD 'RGBQUAD

   'Creates a VB array without copying data
'   With tSA
'       .cbElements = 1
'       .cDims = 2
'       .Bounds(0).lLbound = 0
'       .Bounds(0).cElements = cDIB.height
'       .Bounds(1).lLbound = 0
'       .Bounds(1).cElements = cDIB.BytesPerScanLine
'       .pvData = cDIB.DIBSectionBitsPtr
'   End With
'   CopyMemory VarPtrArray(aDib), VarPtr(tSA), 4
   
   xMax = (cDIB.width - 1) * 3
   Debug.Print "hptr = " & Hex(cDIB.DIBSectionBitsPtr)
   Debug.Print "xMax = " & xMax
   Debug.Print "byte = " & cDIB.BytesPerScanLine
   'Debug.Print "ubound2 = " & UBound(aDib, 2)
   'Debug.Print "ubound1 = " & UBound(aDib, 1)
   
   lngBytesPerScanLine = cDIB.BytesPerScanLine
   lpbitsptr = cDIB.DIBSectionBitsPtr
   
   yMax = cDIB.height - 1
   nDepth = nLevels
   
   ' Update our status display prop
    cDIBSave.LineTotal = cDIB.height
   
   ReDim aNodes(1 To 50) 'Allocates initial storage
   TopGarbage = 0
   cNodes = 1
   cClr = 0
   
   
   For y = 0 To yMax
       For x = 0 To xMax Step 3
         'Adds the current pixel to the color octree
         apiCopyMemory bgr, ByVal lpbitsptr + x + (y * lngBytesPerScanLine), 3
         AddClr 1, 1, 0, 255, 0, 255, 0, 255, _
            bgr.Red, bgr.Green, bgr.Blue
           ' rgb.Blue, rgb.Green, rgb.Red
            
          'AddClr 1, 1, 0, 255, 0, 255, 0, 255, _
          '  aDib(x + 2, y), aDib(x + 1, y), aDib(x, y)
            
         
        pPal = -1
        ' SL we are supporting 2 bit color depth.
        ' Skip reducing nodes if we are in 2 bit color depth
        'If PalSize <> 4 Then
         
            Do While cClr > PalSize
               'Combine the levels to get down to desired palette size
               pPal = CombineNodes
               If pPal = False Then Exit Do
            Loop
            ' DEBUG:SL
            ' Took below out. Was causing a failure
            ' when BitDepth = 2
            If pPal = False Then
               If MsgBox("Error in reducing nodes", vbYesNo) = vbNo _
                 Then Stop
            End If
        
       ' End If
      Next x
   ' Update our status display
   cDIBSave.UpdateStatus y, cDIB.width, cDIB.height, "Create Palette", 1
     
   Next y
   
   Clear
   Debug.Print "Colors: " & cClr
   pPal = 0
   Inc = pPal / cNodes
   'Go through octree and extract colors
   For y = 1 To UBound(aNodes)
      If aNodes(y).bIsLeaf Then
         With aNodes(y)
            pPal = pPal + 1
            Add .vRed / .cClrs, .vGreen / .cClrs, .vBlue / .cClrs
         End With
      End If
   Next y
   
    ' Clear the temporary array descriptor
    ' without destroying the local temporary array.
    'CopyMemory ByVal VarPtrArray(aDib), 0&, 4
   Debug.Print pPal

End Sub

'***************************************************************
'                         --Recursive--
'AddClr--Adds a color to the OctTree palette
'  Will call itself if not in correct level
'
'Inputs:
'  iBranch                       nLevel
'  Branch to look down           Current level(depth) in tree
'
'  vMin(R, G, B)                 vMax(R, G, B)
'  The minimum branch value      The maximum branch value
'
'  R, G, B
'  The Red, Green, and Blue color components
'
'Date Created: 7/12/1999         Brian Schimpf
'
'Modifications:                                     Date:
'***************************************************************
Private Sub AddClr(ByVal iBranch As Long, ByVal nLevel As Long, _
      ByVal vMinR As Byte, ByVal vMaxR As Byte, _
      ByVal vMinG As Byte, ByVal vMaxG As Byte, _
      ByVal vMinB As Byte, ByVal vMaxB As Byte, _
      ByVal r As Byte, ByVal g As Byte, ByVal b As Byte)
        
Dim iR As Byte, iG As Byte, iB As Byte
Dim vMid As Long, iIndex As Long

   
   'Find mid values for colors and decide which path to take
   'Also update max and min values for later call to self
   vMid = vMinR / 2 + vMaxR / 2
   If r > vMid Then
      iR = 1
      vMinR = vMid
   Else
      iR = 0
      vMaxR = vMid
   End If
   
   vMid = vMinG / 2 + vMaxG / 2
   If g > vMid Then
      iG = 1
      vMinG = vMid
   Else
      iG = 0
      vMaxG = vMid
   End If
   
   vMid = vMinB / 2 + vMaxB / 2
   If b > vMid Then
      iB = 1
      vMinB = vMid
   Else
      iB = 0
      vMaxB = vMid
   End If
   
   If aNodes(iBranch).iChildren(iR, iG, iB) = 0 Then  'If no child here then...
      iIndex = GetFreeNode 'Get a new node index
      aNodes(iBranch).iChildren(iR, iG, iB) = iIndex
      aNodes(iBranch).cClrs = aNodes(iBranch).cClrs + 1
      With aNodes(iIndex)   'Clear/set data
         .bIsLeaf = (nLevel = nDepth): .vGreen = 0
         .iNext = 0: .cClrs = 0: .vBlue = 0: .vRed = 0
      End With
   Else  'Has a child here
      iIndex = aNodes(iBranch).iChildren(iR, iG, iB)
   End If
   
   
   If aNodes(iIndex).bIsLeaf Then  'If it is a leaf
      With aNodes(iIndex)
         If .cClrs = 0 Then cClr = cClr + 1
         .cClrs = .cClrs + 1
         .vRed = .vRed + r
         .vGreen = .vGreen + g
         .vBlue = .vBlue + b
      End With
   Else  'Otherwise
      With aNodes(iIndex)
         If .bAddedReduce = False Then    'If 2 or more colors, add to reducable aNodes list
            .iNext = aReduce(nLevel)
            aReduce(nLevel) = iIndex
            .bAddedReduce = True
            'Debug.Print "Reducable Added: " & iIndex, "Level = " & nLevel, "Top was = " & .iNext
         End If
      End With
      
      'Search a level deeper
      AddClr iIndex, nLevel + 1, vMinR, vMaxR, vMinG, vMaxG, _
         vMinB, vMaxB, r, g, b
   End If
End Sub

'***************************************************************
'CombineNodes--Combines octree aNodes to reduce the count of
'  colors.  Combines all children of a leaf into itself.
'
'Date Created: 7/13/1999         Brian Schimpf
'
'Modifications:                                     Date:
'***************************************************************
Private Function CombineNodes() As Boolean
Dim i As Long, iIndex As Long
Dim iR As Byte, iG As Byte, iB As Byte
Dim nR As Long, nG As Long, nB As Long, nPixel As Long
   
   
   
   For i = nDepth To 1 Step -1   'Find deepest reducable level
      If aReduce(i) <> 0 Then Exit For
   Next i
   
   'Debug.Print "aReduce Index = " & i
   
   If i = 0 Then Exit Function
   iIndex = aReduce(i)
   aReduce(i) = aNodes(iIndex).iNext
   
   For i = 0 To 7
      'iR = IIf((i And 1) = 1, 1, 0)
      If (i And 1) = 1 Then
         iR = 1
      Else
         iR = 0
      End If
      'iG = IIf((i And 2) = 2, 1, 0)
      If (i And 2) = 2 Then
         iG = 1
      Else
         iG = 0
      End If
      'iB = IIf((i And 4) = 4, 1, 0)
      If (i And 4) = 4 Then
         iB = 1
      Else
         iB = 0
      End If
      
      If aNodes(iIndex).iChildren(iR, iG, iB) <> 0 Then 'If there is a child
         With aNodes(aNodes(iIndex).iChildren(iR, iG, iB))
            'Add red, green, blue, and pixel count to running total
            nR = nR + .vRed
            nG = nG + .vGreen
            nB = nB + .vBlue
            nPixel = nPixel + .cClrs
            FreeNode aNodes(iIndex).iChildren(iR, iG, iB) 'Free the node
            cClr = cClr - 1
         End With
         aNodes(iIndex).iChildren(iR, iG, iB) = 0 'Clear the link
      End If
   Next i
   
   cClr = cClr + 1
   
   'Set the new node data
   With aNodes(iIndex)
      .cClrs = nPixel
      .bIsLeaf = True
      .vRed = nR
      .vBlue = nB
      .vGreen = nG
   End With
   CombineNodes = True
   
End Function

Public Sub EraseNodes()
  Erase aNodes
  TopGarbage = 0
End Sub

'***************************************************************
'FreeNode--Puts a node on the top of the garbage list
'
'Inputs:
'  iNode
'  Index of node to free
'
'Date Creates:7/13/1999          Brian Schimpf
'
'Modifications:                                     Date:
'***************************************************************
Private Sub FreeNode(ByVal iNode As Long)
   aNodes(iNode).iNext = TopGarbage
   TopGarbage = iNode
   aNodes(iNode).bIsLeaf = False 'Necessary for final loop through
   aNodes(iNode).bAddedReduce = False
   cNodes = cNodes - 1
End Sub

'***************************************************************
'GetFreeNode--Gets a new node index from the trash list or the
'  end of the list.  Clears child pointers
'
'Outputs:
'  Node Index
'
'Date Created:7/12/1999          Brian Schimpf
'
'Modifications:                                     Date:
'***************************************************************
Private Function GetFreeNode() As Long
Dim i As Long, iR As Byte, iG As Byte, iB As Byte
   cNodes = cNodes + 1
   If TopGarbage = 0 Then
      If cNodes > UBound(aNodes) Then
         i = cNodes * 1.1
         ReDim Preserve aNodes(1 To i)
      End If
      GetFreeNode = cNodes
   Else
      GetFreeNode = TopGarbage
      TopGarbage = aNodes(TopGarbage).iNext
      For i = 0 To 7
         'iR = IIf((i And 1) = 1, 1, 0)
         If (i And 1) = 1 Then
            iR = 1
         Else
            iR = 0
         End If
         'iG = IIf((i And 2) = 2, 1, 0)
         If (i And 2) = 2 Then
            iG = 1
         Else
            iG = 0
         End If
         'iB = IIf((i And 4) = 4, 1, 0)
         If (i And 4) = 4 Then
            iB = 1
         Else
            iB = 0
         End If
         aNodes(GetFreeNode).iChildren(iR, iG, iB) = 0
      Next i
   End If
End Function

Public Sub Clear()
   Erase m_tPal
   m_iPalette = 0
End Sub

Public Sub ApplyPalette( _
      ByRef cFrom As cDIBSection, _
      ByRef cTo As cDIBSection, _
      ByRef cDIBSave As cDIBSectionSave, _
      Optional ByVal bDiffuseError As Boolean = True)
     
     
     
Dim pict() As Byte
Dim pict2() As Byte
Dim sa As SAFEARRAY2D
Dim sa2 As SAFEARRAY2D
Dim x As Long, y As Long
Dim i As Long, iCoeff As Long, j As Long
Dim lTIme As Long
Dim xMax As Long, yMax As Long
Dim lErrorRed As Long, lErrorBlue As Long, lErrorGreen As Long
Dim lNewRed As Long, lNewBlue As Long, lNewGreen As Long
Dim lIndex As Long
Dim iC As Long, iC2 As Long
Dim ctr As Long


'Dim lngBytesPerScanLine As Long
Dim lpbitsptrTO As Long
Dim lpbitsptrFROM As Long
Dim rgb As RGBQUAD
Dim bgr As BGRQUAD
' Init our loop counter
ctr = 0
    
    ' have the local matrix point to bitmap pixels
'    With sa
'        .cbElements = 1
'        .cDims = 2
'        .Bounds(0).lLbound = 0
'        .Bounds(0).cElements = cTo.height
'        .Bounds(1).lLbound = 0
'        .Bounds(1).cElements = cTo.BytesPerScanLine
'        .pvData = cTo.DIBSectionBitsPtr
'    End With
'    CopyMemory ByVal VarPtrArray(pict), VarPtr(sa), 4
'    ' Pict now stores the To buffer
'
'    ' have the local matrix point to bitmap pixels
'    With sa2
'        .cbElements = 1
'        .cDims = 2
'        .Bounds(0).lLbound = 0
'        .Bounds(0).cElements = cFrom.height
'        .Bounds(1).lLbound = 0
'        .Bounds(1).cElements = cFrom.BytesPerScanLine
'        .pvData = cFrom.DIBSectionBitsPtr
'    End With
    'CopyMemory ByVal VarPtrArray(pict2), VarPtr(sa2), 4
    ' Pict2 now stores the From buffer
   
    yMax = cTo.height - 1
    xMax = (cTo.width - 1) * 3
    
    
     lngBytesPerScanLine = cFrom.BytesPerScanLine
   lpbitsptrFROM = cFrom.DIBSectionBitsPtr
   lpbitsptrTO = cTo.DIBSectionBitsPtr
    
    ' Update our status display prop
    'cDIBSave.LineTotal = cTo.width
     cDIBSave.LineTotal = cTo.height
    
    
    
    
    For y = 0 To yMax
       For x = 0 To xMax Step 3
    
    
    
    
    ' Get nearest colour:
    'lIndex = ClosestIndex(pict2(x + 2, y), pict2(x + 1, y), pict2(x, y))
    apiCopyMemory bgr, ByVal lpbitsptrFROM + x + (y * lngBytesPerScanLine), 3
    lIndex = ClosestIndex(bgr.Red, bgr.Green, bgr.Blue)
                                   
                                    
            'pict(x + 2, y) = Red(lIndex)
            'pict(x + 1, y) = Green(lIndex)
            'pict(x, y) = Blue(lIndex)
                        
            bgr.Blue = Blue(lIndex)
            bgr.Green = Green(lIndex)
            bgr.Red = Red(lIndex)
            ' Mod sl JUly 6
            'rgb.Red = Red(lIndex)
            'rgb.Green = Green(lIndex)
            'rgb.Blue = Blue(lIndex)
            
            apiCopyMemory ByVal lpbitsptrTO + x + (y * lngBytesPerScanLine), bgr, 3
              
              ' *********************************
              ' ******** DEBUG FIX THIS AFTER!!!!!!!!
              'bDiffuseError = False
              
              
            If bDiffuseError Then
               lErrorBlue = -1 * (CLng(GetByte(lpbitsptrTO, x + 2, y)) - GetByte(lpbitsptrFROM, x + 2, y))
               lErrorGreen = -1 * (CLng(GetByte(lpbitsptrTO, x + 1, y)) - GetByte(lpbitsptrFROM, x + 1, y))
               lErrorRed = -1 * (CLng(GetByte(lpbitsptrTO, x, y)) - GetByte(lpbitsptrFROM, x, y))
               
               'lErrorRed = -1 * (CLng(pict(x + 2, y)) - pict2(x + 2, y))
               'lErrorGreen = -1 * (CLng(pict(x + 1, y)) - pict2(x + 1, y))
               'lErrorBlue = -1 * (CLng(pict(x, y)) - pict2(x, y))
               
               ' Diffuse the error:
               'Debug.Print lErrorRed, lErrorGreen, lErrorBlue
               If Abs(lErrorRed) + Abs(lErrorGreen) + Abs(lErrorBlue) > 3 Then
                  If (x < xMax - 3) Then
                      lNewBlue = GetByte(lpbitsptrFROM, x + 3, y) + (lErrorBlue * 7) \ 16
                      lNewGreen = GetByte(lpbitsptrFROM, x + 4, y) + (lErrorGreen * 7) \ 16
                      lNewRed = GetByte(lpbitsptrFROM, x + 5, y) + (lErrorRed * 7) \ 16
                      
                      'lNewBlue = pict2(x + 3, y) + (lErrorBlue * 7) \ 16
                      'lNewGreen = pict2(x + 4, y) + (lErrorGreen * 7) \ 16
                      'lNewRed = pict2(x + 5, y) + (lErrorRed * 7) \ 16
                      Range lNewBlue, 0, 255
                      Range lNewGreen, 0, 255
                      Range lNewRed, 0, 255
                      'pict2(x + 3, y) = lNewBlue
                      'pict2(x + 4, y) = lNewGreen
                      'pict2(x + 5, y) = lNewRed
                      
                      bgr.Blue = lNewBlue
                      bgr.Green = lNewGreen
                      bgr.Red = lNewRed
                      apiCopyMemory ByVal lpbitsptrFROM + x + 3 + (y * lngBytesPerScanLine), bgr, 3
                      
                      
                      
                      
                  End If
                  If (y < yMax) Then
                      For i = -3 To 3 Step 3
                          If (x + i) > 0 And (x + i) < xMax Then
                              Select Case i
                              Case -3
                                  iCoeff = 0
                              Case 0
                                  iCoeff = 4
                              Case 3
                                  iCoeff = 0
                              End Select
                              lNewBlue = GetByte(lpbitsptrFROM, x + i, y + 1) + (lErrorBlue * iCoeff) \ 16
                              lNewGreen = GetByte(lpbitsptrFROM, x + i + 1, y + 1) + (lErrorGreen * iCoeff) \ 16
                              lNewRed = GetByte(lpbitsptrFROM, x + i + 2, y + 1) + (lErrorRed * iCoeff) \ 16
                              Range lNewBlue, 0, 255
                              Range lNewGreen, 0, 255
                              Range lNewRed, 0, 255
                              'pict2(x + i, y + 1) = lNewBlue
                              'pict2(x + i + 1, y + 1) = lNewGreen
                              'pict2(x + i + 2, y + 1) = lNewRed
                              
                            bgr.Blue = lNewBlue
                            bgr.Green = lNewGreen
                            bgr.Red = lNewRed
                      apiCopyMemory ByVal lpbitsptrFROM + x + i + ((y + 1) * lngBytesPerScanLine), bgr, 3
                      
                              
                              
                              
                              
                          End If
                      Next i
                  End If
               End If
            End If
        Next x ' debug was y
    ' Update our status display
    'ctr = ctr + 1
    'cDIBSave.UpdateStatus ctr, cFrom.width, cFrom.height, "Apply Palette", 2
    
    ' Update our status display
   cDIBSave.UpdateStatus y, cFrom.width, cFrom.height, "Apply Palette", 2
    
    
    Next y  ' *** DEBUG  WAW x
   
   
    ' Clear the temporary array descriptor
    ' without destroying the local temporary array.
    'CopyMemory ByVal VarPtrArray(pict), 0&, 4
    'CopyMemory ByVal VarPtrArray(pict2), 0&, 4
    
   Debug.Print iC, iC2
    
    
End Sub
   
   
Private Function GetByte(lpAdd As Long, x As Long, y As Long) As Long
apiCopyMemory GetByte, ByVal lpAdd + x + (y * lngBytesPerScanLine), 1
   
End Function

   
Private Sub Range( _
      ByRef lIn As Long, _
      ByVal lMin As Long, _
      ByVal lMax As Long _
   )
   If (lIn < lMin) Then
      lIn = lMin
   ElseIf (lIn > lMax) Then
      lIn = lMax
   End If
End Sub


